@@ -2,7 +2,16 @@
    Copyright (C) 2001, 2002, 2003, 2004, 2005
    Craig Franklin
 
-    Copyright (C) 2015-2016 Molnar Karoly
+   Copyright (C) 2015-2016 Molnár Károly
+
+   Dump COFF file contents option.
+   Copyright (C) 2019 Gonzalo Pérez de Olaguer Córdoba <salo@gpoc.es>
+
+   Experimental pagesel and banksel removal.
+   Copyright (C) 2017 Gonzalo Pérez de Olaguer Córdoba <salo@gpoc.es>
+
+   Experimental PCALLW removal.
+   Copyright (C) 2017 Gonzalo Pérez de Olaguer Córdoba <salo@gpoc.es>
 
 This file is part of gputils.
 
@@ -40,39 +49,49 @@ struct gplink_state state;
 static gp_boolean   processor_mismatch_warning;
 static gp_boolean   enable_cinit_wanings;
 
-#define GET_OPTIONS "a:b:cCdf:hI:jlmo:O:p:qrs:t:u:vw"
+#define GET_OPTIONS "a:b:BcCdf:hI:jlmo:O:p:Pqrs:S:t:u:vwW"
 
 enum {
   OPT_MPLINK_COMPATIBLE = 0x100,
   OPT_STRICT_OPTIONS
+#ifdef GPUTILS_DEBUG
+  , OPT_DUMP_COFF
+#endif
 };
 
 static struct option longopts[] =
 {
-  { "hex-format",         required_argument, NULL, 'a' },
-  { "optimize-banksel",   required_argument, NULL, 'b' },
-  { "object",             no_argument,       NULL, 'c' },
-  { "no-cinit-warnings",  no_argument,       NULL, 'C' },
-  { "debug",              no_argument,       NULL, 'd' },
-  { "fill",               required_argument, NULL, 'f' },
-  { "help",               no_argument,       NULL, 'h' },
-  { "include",            required_argument, NULL, 'I' },
-  { "no-save-local",      no_argument,       NULL, 'j' },
-  { "no-list",            no_argument,       NULL, 'l' },
-  { "map",                no_argument,       NULL, 'm' },
-  { "output",             required_argument, NULL, 'o' },
-  { "optimize",           required_argument, NULL, 'O' },
-  { "optimize-pagesel",   required_argument, NULL, 'p' },
-  { "quiet",              no_argument,       NULL, 'q' },
-  { "use-shared",         no_argument,       NULL, 'r' },
-  { "script",             required_argument, NULL, 's' },
-  { "stack",              required_argument, NULL, 't' },
-  { "strict-options",     no_argument,       NULL, OPT_STRICT_OPTIONS },
-  { "macro",              required_argument, NULL, 'u' },
-  { "version",            no_argument,       NULL, 'v' },
-  { "processor-mismatch", no_argument,       NULL, 'w' },
-  { "mplink-compatible",  no_argument,       NULL, OPT_MPLINK_COMPATIBLE },
-  { NULL,                 no_argument,       NULL, '\0'}
+  { "hex-format",           required_argument, NULL, 'a' },
+  { "optimize-banksel",     required_argument, NULL, 'b' },
+  { "experimental-banksel", no_argument,       NULL, 'B' },
+  { "object",               no_argument,       NULL, 'c' },
+  { "no-cinit-warnings",    no_argument,       NULL, 'C' },
+  { "debug",                no_argument,       NULL, 'd' },
+  { "fill",                 required_argument, NULL, 'f' },
+  { "help",                 no_argument,       NULL, 'h' },
+  { "include",              required_argument, NULL, 'I' },
+  { "no-save-local",        no_argument,       NULL, 'j' },
+  { "no-list",              no_argument,       NULL, 'l' },
+  { "map",                  no_argument,       NULL, 'm' },
+  { "output",               required_argument, NULL, 'o' },
+  { "optimize",             required_argument, NULL, 'O' },
+  { "optimize-pagesel",     required_argument, NULL, 'p' },
+  { "experimental-pagesel", no_argument,       NULL, 'P' },
+  { "quiet",                no_argument,       NULL, 'q' },
+  { "use-shared",           no_argument,       NULL, 'r' },
+  { "script",               required_argument, NULL, 's' },
+  { "stack",                required_argument, NULL, 't' },
+  { "strict",               required_argument, NULL, 'S' },
+  { "strict-options",       no_argument,       NULL, OPT_STRICT_OPTIONS },
+  { "macro",                required_argument, NULL, 'u' },
+  { "version",              no_argument,       NULL, 'v' },
+  { "processor-mismatch",   no_argument,       NULL, 'w' },
+  { "experimental-pcallw",  no_argument,       NULL, 'W' },
+  { "mplink-compatible",    no_argument,       NULL, OPT_MPLINK_COMPATIBLE },
+#ifdef GPUTILS_DEBUG
+  { "dump-coff",            no_argument,       NULL, OPT_DUMP_COFF },
+#endif
+  { NULL,                   no_argument,       NULL, '\0'}
 };
 
 /*------------------------------------------------------------------------------------------------*/
@@ -84,10 +103,14 @@ _show_usage(void)
   printf("Options: [defaults in brackets after descriptions]\n");
   printf("  -a FMT, --hex-format FMT       Select hex file format.\n");
   printf("  -b OPT, --optimize-banksel OPT Remove unnecessary Banksel directives. [0]\n");
+  printf("  -B, --experimental-banksel     Use experimental Banksel removal.\n");
   printf("  -c, --object                   Output executable object file.\n");
   printf("  -C, --no-cinit-warnings        Disable this warnings of _cinit section with -O2 option:\n"
          "                                   \"Relocation symbol _cinit has no section.\"\n");
   printf("  -d, --debug                    Output debug messages.\n");
+#ifdef GPUTILS_DEBUG
+  printf("      --dump-coff                Dump COFF file contents.\n");
+#endif
   printf("  -f VALUE, --fill VALUE         Fill unused program memory with value.\n");
   printf("  -h, --help                     Show this usage message.\n");
   printf("  -I DIR, --include DIR          Specify include directory.\n");
@@ -98,15 +121,21 @@ _show_usage(void)
   printf("  -o FILE, --output FILE         Alternate name of output file.\n");
   printf("  -O OPT, --optimize OPT         Optimization level. [1]\n");
   printf("  -p OPT, --optimize-pagesel OPT Remove unnecessary Pagesel directives. [0]\n");
+  printf("  -P, --experimental-pagesel     Use experimental Pagesel removal.\n");
   printf("  -q, --quiet                    Quiet.\n");
   printf("  -r, --use-shared               Use shared memory if necessary.\n");
   printf("  -s FILE, --script FILE         Linker script.\n");
   printf("  -t SIZE, --stack SIZE          Create a stack section.\n");
+  printf("  -S [0|1|2], --strict [0|1|2]   Set the strict level of the missing symbol.\n"
+         "                                     0: This is the default. No message.\n"
+         "                                     1: Show warning message if there is missing symbol.\n"
+         "                                     2: Show error message if there is missing symbol.\n");
   printf("      --strict-options           If this is set, then an option may not be parameter\n"
          "                                   of an another option. For example: -s --quiet\n");
   printf("  -u, --macro symbol[=value]     Add macro value for script.\n");
   printf("  -v, --version                  Show version.\n");
   printf("  -w, --processor-mismatch       Disable \"processor mismatch\" warning.\n");
+  printf("  -W, --experimental-pcallw      Remove unnecessary PCALLW stubs created by SDCC.\n");
   printf("\n");
 #ifdef USE_DEFAULT_PATHS
   if (gp_lkr_path != NULL) {
@@ -382,6 +411,7 @@ _build_tables(void)
   const symbol_t        *sym;
   const char            *name;
   const gp_coffsymbol_t *var;
+  gp_boolean             need_exit;
 
   /* Create the object file symbol tables. */
   object = state.object;
@@ -435,16 +465,29 @@ _build_tables(void)
   /* All of the archives have been scanned. If there are still missing
      references, it is an error. */
   if (_count_missing() > 0) {
+    need_exit = false;
     for (i = 0; i < gp_sym_get_symbol_count(state.symbol.missing); ++i) {
       sym  = gp_sym_get_symbol_with_index(state.symbol.missing, i);
       name = gp_sym_get_symbol_name(sym);
       assert(name != NULL);
       var = (const gp_coffsymbol_t *)gp_sym_get_symbol_annotation(sym);
       assert(var != NULL);
-      gp_error("Missing definition for symbol \"%s\", required by \"%s\".", name, var->file->filename);
+
+      switch (state.strict_level) {
+        case 1:
+          gp_warning("Missing definition for symbol \"%s\", required by \"%s\".", name, var->file->filename);
+          break;
+
+        case 2:
+          gp_error("Missing definition for symbol \"%s\", required by \"%s\".", name, var->file->filename);
+          need_exit = true;
+          break;
+      }
     }
 
-    exit(1);
+    if (need_exit) {
+      exit(1);
+    }
   }
 }
 
@@ -459,19 +502,24 @@ gplink_open_coff(const char *Name)
   gp_object_t  *object;
   gp_archive_t *archive;
   FILE         *coff;
-  char          file_name[PATH_MAX + 1];
+  char         *full_name;
+  int           i;
+  int           len;
 
-  strncpy(file_name, Name, sizeof(file_name));
+  full_name = GP_Strdup(Name);
 
-  coff = fopen(file_name, "rb");
-  if ((coff == NULL) && (strchr(file_name, PATH_SEPARATOR_CHAR) == 0)) {
+  coff = fopen(full_name, "rb");
+  if ((coff == NULL) && (strchr(full_name, PATH_SEPARATOR_CHAR) == NULL)) {
     /* If no PATH_SEPARATOR_CHAR in name, try searching include pathes. */
-    int i;
-
     for (i = 0; i < state.num_paths; i++) {
-      snprintf(file_name, sizeof(file_name), "%s" PATH_SEPARATOR_STR "%s", state.paths[i], Name);
-      coff = fopen(file_name, "rb");
+      len = snprintf(NULL, 0, "%s" PATH_SEPARATOR_STR "%s", state.paths[i], Name);
+      assert(len > 0);
 
+      ++len;
+      full_name = GP_Realloc(full_name, (size_t)len);
+      snprintf(full_name, (size_t)len, "%s" PATH_SEPARATOR_STR "%s", state.paths[i], Name);
+
+      coff = fopen(full_name, "rb");
       if (coff != NULL) {
         break;
       }
@@ -485,32 +533,34 @@ gplink_open_coff(const char *Name)
 
   /* FIXME: Three files are opened, surely one is sufficent. */
 
-  switch (gp_identify_coff_file(file_name)) {
+  switch (gp_identify_coff_file(full_name)) {
     case GP_COFF_OBJECT_V2:
     case GP_COFF_OBJECT:
       /* read the object */
-      object = gp_read_coff(file_name);
-      /*object_append(object, file_name);*/
+      object = gp_read_coff(full_name);
+      /*object_append(object, full_name);*/
       _object_append(object);
       break;
 
     case GP_COFF_ARCHIVE:
       /* read the archive */
-      archive = gp_archive_read(file_name);
-      _archive_append(archive, file_name);
+      archive = gp_archive_read(full_name);
+      _archive_append(archive, full_name);
       break;
 
     case GP_COFF_SYS_ERR:
-      gp_error("Can't open file \"%s\".", file_name);
+      gp_error("Can't open file \"%s\".", full_name);
       break;
 
     case GP_COFF_UNKNOWN:
-      gp_error("\"%s\" is not a valid coff object or archive.", file_name);
+      gp_error("\"%s\" is not a valid coff object or archive.", full_name);
       break;
 
     default:
       assert(0);
   }
+
+  free(full_name);
 }
 
 /*------------------------------------------------------------------------------------------------*/
@@ -557,6 +607,7 @@ _init(void)
   state.lst_file             = OUT_NORMAL;
   state.map_file             = OUT_SUPPRESS;
   state.obj_file             = OUT_SUPPRESS;
+  state.strict_level         = 0;
 
   /* set default output filename to be a.o, a.hex, a.cod, a.map */
   strncpy(state.base_file_name, "a", sizeof(state.base_file_name));
@@ -685,18 +736,26 @@ _process_args(int Argc, char *Argv[])
         break;
       }
 
-      case 'c':
+      case 'B': {
+        state.optimize.experimental_banksel = true;
+        break;
+      }
+
+      case 'c': {
         state.obj_file = OUT_NORMAL;
         break;
+      }
 
-      case 'C':
+      case 'C': {
         enable_cinit_wanings = false;
         break;
+      }
 
-      case 'd':
+      case 'd': {
         gp_debug_disable = false;
         yydebug = 1;
         break;
+      }
 
       case 'f': {
         state.fill_value = strtol(optarg, &pc, 16);
@@ -714,25 +773,30 @@ _process_args(int Argc, char *Argv[])
       }
 
       case '?':
-      case 'h':
+      case 'h': {
         usage = true;
         break;
+      }
 
-      case 'I':
+      case 'I': {
         gplink_add_path(optarg);
         break;
+      }
 
-      case 'j':
+      case 'j': {
         state.cod.no_save_local = true;
         break;
+      }
 
-      case 'l':
+      case 'l': {
         state.lst_file = OUT_SUPPRESS;
         break;
+      }
 
-      case 'm':
+      case 'm': {
         state.map_file = OUT_NORMAL;
         break;
+      }
 
       case 'o': {
         strncpy(state.base_file_name, optarg, sizeof(state.base_file_name));
@@ -744,9 +808,10 @@ _process_args(int Argc, char *Argv[])
         break;
       }
 
-      case 'O':
+      case 'O': {
         /* do nothing */
         break;
+      }
 
       case 'p': {
         state.optimize.pagesel = (unsigned int)strtol(optarg, &pc, 10);
@@ -757,13 +822,20 @@ _process_args(int Argc, char *Argv[])
         break;
       }
 
-      case 'q':
+      case 'P': {
+        state.optimize.experimental_pagesel = true;
+        break;
+      }
+
+      case 'q': {
         gp_quiet = true;
         break;
+      }
 
-      case 'r':
+      case 'r': {
         gp_relocate_to_shared = true;
         break;
+      }
 
       case 's': {
         fn = GP_Malloc(sizeof(srcfns_t));
@@ -781,6 +853,15 @@ _process_args(int Argc, char *Argv[])
         break;
       }
 
+      case 'S': {
+        state.strict_level = atoi(optarg);
+
+        if ((state.strict_level < 0) || (state.strict_level > 2)) {
+          gp_error("Invalid strict level: %d (Must be: 0, 1, 2)", state.strict_level);
+        }
+        break;
+      }
+
       case 't': {
         state.stack_size = strtol(optarg, &pc, 10);
 
@@ -794,26 +875,43 @@ _process_args(int Argc, char *Argv[])
         break;
       }
 
-      case 'u':
+      case 'u': {
         _parse_define(optarg, script_add_symbol_value);
         break;
+      }
 
-      case 'v':
+      case 'v': {
         fprintf(stderr, "%s\n", GPLINK_VERSION_STRING);
         exit(0);
         break;
+      }
 
-      case 'w':
+      case 'w': {
         processor_mismatch_warning = false;
         break;
+      }
 
-      case OPT_MPLINK_COMPATIBLE:
+      case 'W': {
+        state.optimize.experimental_pcallw = true;
+        break;
+      }
+
+      case OPT_MPLINK_COMPATIBLE: {
         state.mplink_compatible = true;
         break;
+      }
 
-      case OPT_STRICT_OPTIONS:
+#ifdef GPUTILS_DEBUG
+      case OPT_DUMP_COFF: {
+        gp_dump_coff = true;
+        break;
+      }
+#endif
+
+      case OPT_STRICT_OPTIONS: {
         /* do nothing */
         break;
+      }
     } /* switch (c) */
 
     if (usage) {
@@ -826,8 +924,6 @@ _process_args(int Argc, char *Argv[])
     pc = strrchr(Argv[optind], '.');
 
     if ((pc != NULL) && (strcasecmp(pc, ".lkr") == 0)) {
-      srcfns_t *fn;
-
       fn = GP_Malloc(sizeof(srcfns_t));
       fn->file_name = GP_Strdup(Argv[optind++]);
       fn->next      = NULL;
@@ -1020,12 +1116,46 @@ _linker(void)
 
   gp_cofflink_update_table(state.object, state.class->org_to_byte_shift);
 
-  if (state.optimize.pagesel > 0) {
-    gp_coffopt_remove_unnecessary_pagesel(state.object);
+  /* remove unnecessary pcallw stubs (created by the SDCC compiler)
+   *
+   * current implementation requires this to be done after
+   * sections have been relocated.
+   */
+  if (state.optimize.experimental_pcallw) {
+    gp_coffopt_remove_unnecessary_pcallw_experimental(state.object);
   }
 
-  if (state.optimize.banksel > 0) {
-    gp_coffopt_remove_unnecessary_banksel(state.object);
+  /* after pagesel removal has been done sections cannot be moved
+   * across pages, so it is better to do pagesel removal last.
+   */
+
+  if (state.optimize.experimental_banksel || state.optimize.experimental_pagesel) {
+    if (state.optimize.banksel > 0) {
+      if (state.optimize.experimental_banksel) {
+        gp_coffopt_remove_unnecessary_banksel_experimental(state.object);
+      }
+      else {
+        gp_coffopt_remove_unnecessary_banksel(state.object);
+      }
+    }
+
+    if (state.optimize.pagesel > 0) {
+      if (state.optimize.experimental_pagesel) {
+        gp_coffopt_remove_unnecessary_pagesel_experimental(state.object);
+      }
+      else {
+        gp_coffopt_remove_unnecessary_pagesel(state.object);
+      }
+    }
+  }
+  else {
+    if (state.optimize.pagesel > 0) {
+      gp_coffopt_remove_unnecessary_pagesel(state.object);
+    }
+
+    if (state.optimize.banksel > 0) {
+      gp_coffopt_remove_unnecessary_banksel(state.object);
+    }
   }
 
   gp_coffgen_make_linenum_array(state.object);
